"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const apollo_server_express_1 = require("apollo-server-express");
const cors_1 = __importDefault(require("cors"));
const express_1 = __importDefault(require("express"));
const graphql_depth_limit_1 = __importDefault(require("graphql-depth-limit"));
const typeorm_1 = require("typeorm");
const database_connector_1 = __importDefault(require("../connectors/database-connector"));
const redis_connector_1 = __importDefault(require("../connectors/redis-connector"));
const loaders_1 = require("../loaders");
const rest_loader_1 = __importDefault(require("../loaders/rest.loader"));
const logger_1 = __importDefault(require("../logger"));
const tool_1 = require("../tool");
const auth_directive_1 = __importDefault(require("./auth.directive"));
class App {
    constructor() {
        this.gqlLoader = new loaders_1.GqlLoader();
        this.expressApp = express_1.default();
        this.initContext = {};
    }
    static playgroundUrl(port) {
        return tool_1.EnvTool.endpoint ? `${tool_1.EnvTool.endpoint}/graphql` : `http://localhost:${port || 3000}/graphql`;
    }
    async loadConfig() {
        var _a, _b, _c;
        const config = await tool_1.YamlParser.parse(tool_1.PathsResolverTool.YAML);
        if (tool_1.EnvTool.launcher) {
            config.localConfiguration.database = tool_1.EnvTool.dbConfig;
        }
        else {
            config.localConfiguration.database.synchronize = true;
        }
        this.config = config;
        this.expressPort = tool_1.EnvTool.launcher ? 3000 : ((_c = (_b = (_a = this.config) === null || _a === void 0 ? void 0 : _a.localConfiguration) === null || _b === void 0 ? void 0 : _b.application) === null || _c === void 0 ? void 0 : _c.port) || 3000;
    }
    async buildServer(initContextStorage) {
        const context = 'Bootstrap';
        try {
            await this.loadConfig();
            const dbConnection = await this.createDbConnectionAndMigrate();
            const { redis, platformRedis } = await this.createRedisConnections();
            this.initContext = await App.getInitContext();
            this.expressApp.use(cors_1.default());
            await this.buildGraphql(dbConnection, redis, platformRedis);
            const endpoints = await this.buildRest(dbConnection, redis);
            this.expressApp.use('/health', (_, res) => res.json());
            if (initContextStorage) {
                initContextStorage(this.expressApp);
            }
            let server;
            await new Promise((resolve) => {
                server = this.expressApp.listen(this.expressPort, () => {
                    logger_1.default.log(`GraphQL is available on: http://localhost:${this.expressPort}/graphql`, context);
                    if (endpoints.length) {
                        logger_1.default.log('REST endpoints are available on:', context);
                        const endpointsFlatten = {};
                        endpoints.forEach((endpoint) => {
                            (endpointsFlatten[endpoint.route] = endpointsFlatten[endpoint.route] || []).push(endpoint.method);
                        });
                        Object.entries(endpointsFlatten).forEach((endpoint) => {
                            logger_1.default.log(`  -> [${endpoint[1].join(', ').toUpperCase()}] http://localhost:${this.expressPort}/rest${endpoint[0]}`, context);
                        });
                    }
                    resolve();
                });
            });
            server.setTimeout(60000);
            return server.close.bind(server);
        }
        catch (e) {
            throw new Error(`Server failed to start, error: ${e.message}`);
        }
    }
    async buildGraphql(connection, redis, platformRedis) {
        var _a, _b;
        const contextHandler = await this.gqlLoader.getContextHandler();
        const requestCostTool = new tool_1.RequestCostTool(platformRedis);
        const costAnalysis = await tool_1.costAnalysisTool(this.config, requestCostTool);
        const schema = await loaders_1.GqlLoader.getTypeDefs();
        const server = new apollo_server_express_1.ApolloServer({
            playground: {
                endpoint: App.playgroundUrl(this.expressPort),
            },
            introspection: true,
            typeDefs: schema,
            resolvers: Object.assign(Object.assign({}, await this.gqlLoader.getResolvers()), this.gqlLoader.customScalars(schema)),
            uploads: false,
            schemaDirectives: {
                auth: auth_directive_1.default,
            },
            validationRules: [
                graphql_depth_limit_1.default(((_b = (_a = this.config.security) === null || _a === void 0 ? void 0 : _a.graphql) === null || _b === void 0 ? void 0 : _b.maxDepth) || 10),
            ],
            plugins: [
                costAnalysis.plugin,
            ],
            context: async (expressContext) => {
                const context = {
                    initContext: this.initContext,
                    db: { connection },
                    redis,
                    request: expressContext.req,
                };
                if (contextHandler && expressContext.req.body.operationName !== 'IntrospectionQuery') {
                    const result = await contextHandler(context);
                    if (result && typeof result === 'object') {
                        return Object.assign(Object.assign({}, context), result);
                    }
                }
                return context;
            },
        });
        this.expressApp.use('/graphql', express_1.default.json(), costAnalysis.middleware);
        server.applyMiddleware({ app: this.expressApp });
    }
    async buildRest(connection, redis) {
        return new rest_loader_1.default(this.expressApp, { db: { connection }, initContext: this.initContext, redis }, this.config.rest).loadEndpoints();
    }
    async createDbConnectionAndMigrate() {
        var _a;
        const context = 'Database';
        const { database } = this.config.localConfiguration;
        if ((_a = this.config.serviceConfiguration) === null || _a === void 0 ? void 0 : _a.skipDatabase) {
            logger_1.default.log('Skipping connection to database, to disable - remove skipDatabase flag from codestore.yaml', context);
            return undefined;
        }
        let connection;
        try {
            connection = typeorm_1.getConnectionManager().get('default');
        }
        catch (error) {
            if (error.name === 'ConnectionNotFoundError') {
                logger_1.default.log('Connecting to database', context);
                connection = await database_connector_1.default.createConnection(database);
                logger_1.default.log('Successfully connected', context);
            }
            else
                throw error;
        }
        if (!database.synchronize) {
            await connection.runMigrations();
            logger_1.default.log('Migrations ran', context);
        }
        return connection;
    }
    static async getInitContext() {
        if (await tool_1.PathsResolverTool.resolvePath(tool_1.PathsResolverTool.INDEX)) {
            const { default: defaultExport } = await tool_1.importUncached(tool_1.PathsResolverTool.INDEX);
            logger_1.default.log('Importing initial context from src/index.ts', 'Bootstrap');
            if (typeof defaultExport === 'function') {
                return defaultExport();
            }
            return defaultExport;
        }
        logger_1.default.log('Skipping import to initial context', 'Bootstrap');
        return {};
    }
    async createRedisConnections() {
        const result = {
            platformRedis: await this.connectToPlatformRedis(),
            redis: await this.connectToRedis(),
        };
        if (result.platformRedis) {
            await result.platformRedis.select(15);
        }
        return result;
    }
    connectToPlatformRedis() {
        var _a;
        if (tool_1.EnvTool.launcher) {
            return redis_connector_1.default.createConnection(tool_1.EnvTool.internalRedisConfig);
        }
        const redis = (_a = this.config.localConfiguration) === null || _a === void 0 ? void 0 : _a.redis;
        if (!redis) {
            return undefined;
        }
        return redis_connector_1.default.createConnection(redis);
    }
    async connectToRedis() {
        var _a, _b;
        const context = 'Redis';
        if (!((_a = this.config.serviceConfiguration) === null || _a === void 0 ? void 0 : _a.enableRedis)) {
            logger_1.default.log('Skipping connection to Redis. Set `serviceConfiguration.enableRedis: true` in codestore.yaml to create connection', context);
            return undefined;
        }
        if (tool_1.EnvTool.launcher) {
            return redis_connector_1.default.createConnection(tool_1.EnvTool.redisConfig);
        }
        const redis = (_b = this.config.localConfiguration) === null || _b === void 0 ? void 0 : _b.redis;
        if (!redis) {
            logger_1.default.log('Skipping local connection to Redis.'
                + 'Define `localConfiguration.redis` object in codestore.yaml to create local connection', context);
            return undefined;
        }
        logger_1.default.log('Connecting to Redis', context);
        const connection = await redis_connector_1.default.createConnection(redis);
        if (connection) {
            logger_1.default.log('Successfully connected', context);
        }
        return connection;
    }
}
exports.default = App;
//# sourceMappingURL=app.js.map